#include <math.h>
#include <algorithm>
#include <numeric>

#include <boost/numeric/ublas/io.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>

#include "indices.h"

const unsigned int nchan = 4;
const unsigned int ninds = 10;                // Restoration: based on fixed number of scint.indices:

const double min_flux = 2.0;

namespace indices {
    double microexposure = 1.0;             // in milliseconds
    vector<double> deadtime = scalar_vector<double>( nchan, 20e-6 );
    vector<double> nonpois  = scalar_vector<double>( nchan, 1.0 );
    vector<double> scatter  = scalar_vector<double>( nchan, 0.0 );
    vector<double> backgrd  = scalar_vector<double>( nchan, 0.0 );
    std::vector< vector<double> > flux;
    std::vector< vector<double> > inds;
    std::vector< vector<double> > desi;
    vector<datum> result_flux;
    vector<datum> result_indx;
    vector<datum> result_desi;
}

void indices::calculation( const std::string& s ) {
    std::istringstream  istr( s );
    vector<double> mean( nchan );
    vector<double> vars( ninds );
    vector<double> rho1( ninds );
    vector<double> rho2( nchan );
    istr >> mean >> vars >> rho1 >> rho2;
    if( istr.eof() || istr.fail() ) throw std::ios_base::failure( "Moments input" );
                                // Input filtering for star escape
                                // Re-normalization
    vector<double> renorm = mass_order_prod( mean );
    vars = element_prod( vars, renorm );
    rho1 = element_prod( rho1, renorm );
    rho2 = element_prod( rho2, subrange( renorm, 0, nchan ) );
                                // Poisson and nonlinearity correction
    vector<double> tauxi = element_prod( mean, deadtime )/( -1.0*microexposure );
    vector<double> wtxi  = apply_to_all( tauxi, functor::lambertW<double>() );
    subrange( vars, 0, nchan ) -= element_prod( element_prod( mean, (scalar_vector<double>( nchan, 1.0 ) + tauxi*2.0) ), nonpois );
    mean = element_prod( mean, apply_to_all( wtxi*(-1.0), functor::exp<double>() ) );
    mean = element_prod( ( scalar_vector<double>( nchan, 1.0 ) - scatter), mean )- backgrd*microexposure;
    if( *min_element( mean.begin(), mean.end() ) < min_flux ) throw std::ios_base::failure( "Non star's data" );
    flux.push_back( mean );
    vector<double> corr = apply_to_all( wtxi, functor::exp<double>() ) + tauxi;
    corr = apply_to_all( corr, functor::inv<double>() );
    vector<double> factor = mass_order_prod( corr );
    vars = element_prod( vars, factor );
    rho1 = element_prod( rho1, factor );
    rho2 = element_prod( rho2, subrange( factor, 0, nchan ) );
                                // Indices calculation
    corr = apply_to_all( mean, functor::inv<double>() );
    factor = mass_order_prod( corr );
    vars = element_prod( vars*1.25 - rho1*0.25, factor );
    desi.push_back( (  subrange( vars, 0, nchan )*3.0 +  element_prod( rho2 - subrange( rho1, 0, nchan )*4.0 , subrange( factor, 0, nchan ) ) )/4.5 );
    vars = apply_to_all( vars + scalar_vector<double>( nchan, 1.0 ),  functor::log<double>() );
    vector<double> sigm =  subrange( vars, 0, nchan );
    subrange( vars, nchan, ninds ) =  subrange( mass_order_summ( sigm ), nchan, ninds ) - subrange( vars, nchan, ninds )*2.0;
    inds.push_back( vars );
}

void indices::extract( matrix<double>& fluxess, matrix<double>& indices, matrix<double>& indesis ) {
    unsigned int numb = flux.size();
    if( numb == 0 ) return;
    indices.resize( numb, ninds );
    indesis.resize( numb, nchan );
    fluxess.resize( numb, nchan );
    for( unsigned int i = 0; i < numb; i++ ) row( fluxess, i ) = flux.at(i);
    result( fluxess, result_flux );
    flux.clear();
    for( unsigned int i = 0; i < numb; i++ ) row( indices, i ) = inds.at(i);
    result( indices, result_indx );
    inds.clear();
    for( unsigned int i = 0; i < numb; i++ ) row( indesis, i ) = desi.at(i);
    result( indesis, result_desi );
    desi.clear();
}

void indices::result( const matrix< double >& inp, vector<datum>& out ) {
    unsigned int nrow = inp.size1();
    unsigned int ncol = inp.size2();
    out.resize( ncol );
    double n = (double)nrow;
    if( nrow < 3 ) throw std::ios_base::failure( "Too few (<3) measurements" );
//    if( nrow < 2 ) return;
    for( unsigned int i = 0; i < ncol; i++ ) out(i).mean = sum(column( inp,i ))/n;
    for( unsigned int i = 0; i < ncol; i++) {
        double vars = inner_prod( column( inp, i ), column( inp, i ))/n - out(i).mean*out(i).mean;
        out(i).errs = sqrt( vars/(n-1) )/out(i).mean;
        if( isnan(out(i).errs) || isinf(out(i).errs)  ) {
            std::cerr << "Incorrect error [" << (int)n << "] " << out(i).errs << std::endl;
            out(i).errs = 0.0;
        }
    }
}

std::string indices::out_fluxes( const std::string& stamp ) {
    std::ostringstream str;
    str << "F " <<  stamp << ' ' << std::fixed << result_flux(0) << ' ' << result_flux(1) << ' ' << result_flux(2) << ' ' << result_flux(3);
    return str.str();
}

std::string indices::out_indices( const std::string& stamp ) {
    std::ostringstream str;
    str << "I " <<  stamp << std::indic << std::fixed << result_indx(0) << result_indx(1) << result_indx(2) << result_indx(3)
        << result_indx(4) << result_indx(5) << result_indx(6) << result_indx(7) << result_indx(8)
        << result_indx(9) << result_desi(0) << result_desi(1) << result_desi(2) << result_desi(3);
    return str.str();
}

void indices::set_background( const vector<double>& vals  ) {
    assert( backgrd.size() == vals.size() );
    backgrd = vals;
}

void indices::set_microexposure( double mex ) {
    microexposure = mex;
}

void indices::set_counters( const vector<double>& nl, const vector<double>& np, const vector<double>& sc ) {
    assert( deadtime.size() == nl.size() );
    assert( nonpois.size() == np.size() );
    assert( scatter.size() == sc.size() );

    deadtime = nl*1e-6;
    nonpois = np;
    scatter = sc;
}

const double zcorr[10][10] = {
/*A*/  {-0.010,-0.107, 0.098, 0.102, 0.230, 0.349,-0.268, 0.041,-0.302,-0.042},
/*B*/  {-0.037,-0.121, 0.115, 0.117, 0.241, 0.339,-0.353, 0.031,-0.362,-0.054},
/*C*/  {-0.102,-0.240, 0.024, 0.067, 0.293, 0.480,-0.271, 0.068,-0.347,-0.050},
/*D*/  {-0.175,-0.152,-0.171,-0.114,-0.054,-0.058, 0.088,-0.001, 0.077, 0.004},
/*AB*/ { 0.324, 0.615, 0.023,-0.100,-0.631,-1.034, 0.596,-0.146, 0.760, 0.114},
/*AC*/ { 0.394, 0.685, 0.393, 0.185,-0.384,-0.839,-0.142,-0.183, 0.164, 0.018},
/*AD*/ { 0.374, 0.402, 0.596, 0.410, 0.182, 0.057,-0.711,-0.065,-0.524,-0.075},
/*BC*/ { 0.408, 0.597, 0.514, 0.302,-0.146,-0.479,-0.422,-0.141,-0.149,-0.025},
/*BD*/ { 0.371, 0.321, 0.654, 0.475, 0.347, 0.320,-0.869,-0.029,-0.720,-0.101},
/*CD*/ { 0.400,-0.012, 0.756, 0.633, 0.947, 1.395,-1.133, 0.155,-1.239,-0.156}
};

void indices::desaturation( matrix<double>& indices ) {
    vector<double> meanidx( ninds );
    vector<double> ecorr( ninds );
    vector<double> vcorr( ninds );
    unsigned int nrow = indices.size1();
    for( unsigned int i = 0; i < ninds; i++ ) meanidx(i) = sum( column( indices, i ) )/nrow;
    for( unsigned int i = 0; i < ninds; i++ ) {
        for( unsigned int j = 0; j < ninds; j++ ) ecorr(j) = zcorr[i][j] ;
        vcorr(i) = 1.0/(1.0+1.0*inner_prod( ecorr, meanidx ));
    }
    for( unsigned int i = 0; i < nrow; i++ ) {
        row( indices, i ) = element_prod( row( indices, i ), vcorr );
    }
}


namespace std {
    static const int iword_index = std::ios_base::xalloc();
    std::ostream& indic( std::ostream& stm ){
        stm.iword(iword_index) = true;
        return stm;
    }
    std::ostream& operator<<( std::ostream& stm, const datum& m ){
        int w = ( stm.flags() & std::ios::scientific )? 8 : 5;
        w = ( stm.iword(iword_index) )? w + 2 : w;
        int p = ( stm.iword(iword_index) )? 4 : std::max( 0, 4 - max( 1, (int)ceil(std::log10( m.mean ) ) ) );
        stm << " " << std::setprecision(p) << std::setw(w) << m.mean << " " << std::fixed
            << std::setprecision(3) << std::setw(7) << m.errs;
        return stm;
    }
}
