#include <math.h>
#include <string>
#include <iostream>
#include <sstream>
#include <iomanip>
#include <map>

#include <boost/program_options.hpp>

#include "astrotime.h"
#include "coordinates.h"
#include "parameters_set.h"
#include "indices.h"
#include "profile.h"
#include "weif.h"

namespace po = boost::program_options;

//using namespace boost::numeric::ublas;

const unsigned int nchan=4;
// FIXME: use boost::filesystem for filepaths instead of pure std::string
//const std::string profile_set_dfl = "0.5,1.0,2.0,4.0,8.0,16.0";
//const std::string dataset_dir_dfl = "./";
const std::string weights_dir_dfl = "weights/";
const std::string spectra_dir_dfl = "spectra/";
const std::string response_file_dfl = "mass_maid.crv";
const std::string ccd_response_dfl = "ccd.crv";

template<typename T> T& get( std::map<T,T>& ps, const T& k ){
    std::map<std::string,std::string>::iterator pos = ps.find( k );
    if( pos == ps.end() || pos->second.empty() ) throw std::runtime_error( "Parameter " + k + " missed" );
    return pos->second;
}

namespace parameters {
	struct profile_grid {
		public:
			// FIXME: use initialization list from C++0x
			profile_grid() {
				v.push_back(0.5);
				v.push_back(1.0);
				v.push_back(2.0);
				v.push_back(4.0);
				v.push_back(8.0);
				v.push_back(16.0);
			}
			profile_grid( const std::vector<double>& v ) : v(v) {}
			std::vector<double> v;
	};

	void validate(boost::any& v, const std::vector<std::string>& values, parameters::profile_grid* target_type, int) {
			std::string s = po::validators::get_single_string(values);
			//FIXME: use regexp and validation instead of dumping
			std::replace( s.begin(), s.end(), ',', ' ' );
			v = boost::any(parameters::profile_grid(parameters::get< std::vector<double> >( s )));
			// NOTE: throw po::validation_error("invalid value");
	}
	
    po::variables_map vm;
    double file_accumtime = 0;
    double cmd_accumtime = 0;
    bool   cmd_empiric = false;
    std::vector<double> cmd_grid;
    std::map<std::string,std::string> rPs;
    Spectral_data spectrum_set( response_file_dfl, weights_dir_dfl, spectra_dir_dfl, ccd_response_dfl);
	profile_grid profile_set;
}

Spectral_data& parameters::spectral_data_files() {
    return spectrum_set;
}

double parameters::accumtime() {
    return ( cmd_accumtime != 0 )? cmd_accumtime : file_accumtime; 
}

bool parameters::correction() {
    return cmd_empiric;
}

Geo* parameters::coordinates() {
    static std::string lon_key("general\\site\\longitude");
    static std::string lat_key("general\\site\\latitude");
    static Geo* site = new Geo( Hour( ::get( rPs, lon_key ) ), Degree( ::get( rPs, lat_key ) ) );   //      ??
    return site;
}

int parameters::comand_line_parse( int argc, char **argv ){
    po::options_description desc("Allowed options");
    desc.add_options()
        ( "help,h",         "produce help message" )
        ( "accum-time,a",   po::value<double>(&cmd_accumtime)->default_value(cmd_accumtime),      "force accumulation time" )
        ( "empiric,e",      "indices saturation empiric correction" )
		( "profile-grid,Z", po::value<profile_grid>(&profile_set), "set for profile grid generator" )
        ( "response,r",     po::value<std::string>(&spectrum_set.response)->default_value(response_file_dfl), "file with detector response" )
        ( "ccd-response,d", po::value<std::string>(&spectrum_set.ccdresponse)->default_value(ccd_response_dfl), "file with detector response" )
        ( "weights-dir,w",  po::value<std::string>(&spectrum_set.weights)->default_value(weights_dir_dfl), "weight-files directory" )
        ( "spectra-dir,s",  po::value<std::string>(&spectrum_set.spectra)->default_value(spectra_dir_dfl), "spectra-files directory" );
    try {
        po::store( po::parse_command_line( argc, argv, desc), vm );
        po::notify( vm );
        if( vm.count("help") ) {
            std::cerr << desc << std::endl;
            return -1;
        }
        if( vm.count("empiric") ) cmd_empiric = true;
    } catch( const po::error& e ) {
        std::cerr << e.what() << std::endl;
        std::cerr << desc << std::endl;
        return -1;
    }
    // FIXME: too sophisticated setzcn2 function API
    restoration::setzcn2( wf::grid_generate<double>( profile_set.v.size(), &profile_set.v[0] ) );
//    spectrum_set.root( dataset_dir );
    return 0;
}

void parameters::P_line_parse( const std::string& p, const std::string& s ){
    std::string::size_type i = 0;
    std::string params;
    std::string value;
    if( (i =  s.find( "=", 0 )) != std::string::npos ) {
        params.resize(i);
        std::transform( s.begin(), s.begin()+i, params.begin(), std::ptr_fun(::tolower) );
        value = s.substr(i+1);
        if( params == "operations\\normal mode\\accumtime" ) file_accumtime = get<double>( value );
        if( params == "operations\\normal mode\\exposition" ) indices::set_microexposure( get<double>( value ) );
        if( params == "operations\\background measurement\\background" ){
            std::replace( value.begin(), value.end(), ',', ' ' );
            indices::set_background( get< vector<double>, nchan >( value ) );
        }
        std::map<std::string,std::string>::iterator pos = rPs.find( params );
        if( pos != rPs.end() ) {
            rPs[params] = value;
            std::cout << p << s << std::endl;
        }
    }
}

bool parameters::apertures( vector<double>& innerDiam, vector<double>& outerDiam  ) {
    static std::string magn_key("optics\\common\\magnification");
    static std::string segm_label = "segmentator\\channel ";
    static std::string chan_name = "abcd";
    double magnification = 0.1*get<double>( ::get( rPs, magn_key ) );
    std::string s1;
    std::string s2;
    for( unsigned int i = 0; i < chan_name.size(); i++ ) s1 += " " + ::get( rPs, segm_label + chan_name[i] + "\\inner" );
    innerDiam =  get< vector<double>, nchan >( s1 )*magnification;
    for( unsigned int i = 0; i < chan_name.size(); i++ ) s2 += " " + ::get( rPs, segm_label + chan_name[i] + "\\outer" );
    outerDiam =  get< vector<double>, nchan >( s2 )*magnification;
    return true;
}

bool parameters::counters() {
    static std::string chan_label = "channels\\channel";
    static std::string chan_name = "abcd";
    std::string s1;
    std::string s2;
    std::string s3;
    for( unsigned int i = 0; i < chan_name.size(); i++ ) s1 += " " + ::get( rPs, chan_label + chan_name[i] + "\\nonlinearity" );
    for( unsigned int i = 0; i < chan_name.size(); i++ ) s2 += " " + ::get( rPs, chan_label + chan_name[i] + "\\nonpoisson" );
    for( unsigned int i = 0; i < chan_name.size(); i++ ) s3 += " " + ::get( rPs, chan_label + chan_name[i] + "\\scatter" );
    indices::set_counters( get< vector<double>, nchan >( s1 ), get< vector<double>, nchan >( s2 ), get< vector<double>, nchan >( s3 ) );
    return true;
}

void parameters::init_keys(){
                        //  common_header 
                        //   '='   -   
                        //     
                        //       
    rPs["general\\site\\longitude"] = "";
    rPs["general\\site\\latitude"] = "";
    rPs["operations\\normal mode\\accumtime"] = "";
    rPs["optics\\common\\magnification"] = "";
    rPs["segmentator\\channel a\\inner"] = "";
    rPs["segmentator\\channel b\\inner"] = "";
    rPs["segmentator\\channel c\\inner"] = "";
    rPs["segmentator\\channel d\\inner"] = "";
    rPs["segmentator\\channel a\\outer"] = "";
    rPs["segmentator\\channel b\\outer"] = "";
    rPs["segmentator\\channel c\\outer"] = "";
    rPs["segmentator\\channel d\\outer"] = "";
    rPs["operations\\background measurement\\background"] = "";
    rPs["operations\\normal mode\\exposition"] = "";
    rPs["channels\\channela\\nonlinearity"] = "";
    rPs["channels\\channelb\\nonlinearity"] = "";
    rPs["channels\\channelc\\nonlinearity"] = "";
    rPs["channels\\channeld\\nonlinearity"] = "";
    rPs["channels\\channela\\nonpoisson"] = "";
    rPs["channels\\channelb\\nonpoisson"] = "";
    rPs["channels\\channelc\\nonpoisson"] = "";
    rPs["channels\\channeld\\nonpoisson"] = "";
    rPs["channels\\channela\\scatter"] = "0";
    rPs["channels\\channelb\\scatter"] = "0";
    rPs["channels\\channelc\\scatter"] = "0";
    rPs["channels\\channeld\\scatter"] = "0";

}
