#include <math.h>
#include <string>
#include <iostream>
#include <sstream>
#include <iomanip>
#include <map>
#include <cmath>

#include "indices.h"
#include "integrals.h"
#include "parameters_set.h"
#include "processing.h"
#include "profile.h"
#include "weif.h"

using namespace boost::numeric::ublas;

const int nchan = 4;
const double basis_set[] = { 0.0, 36.0, 0.25, 0.04 };
const double test_set[4] = { 0.02, 30.0, 0.06, 0.01 };

namespace processing {
    Coord<equ> star;
    std::string spectral_type;
    vector<double> basis_grid  =  wf::grid_generate<double>( sizeof(basis_set)/sizeof(basis_set[0]), basis_set );
    Weights basis_functions( nchan, basis_grid );
                        // TEST
    const double power = 1e-13;
    int loop = 0;
    vector<double> hs = wf::grid_generate<double>( sizeof(test_set)/sizeof(test_set[0]), test_set );
    Weights test_functions( nchan, hs );
                        // TEST END
}

unsigned int processing::run( Date& dc, Time& ut, double mz, double moment ) {
    static bool channels_are_set = parameters::counters();
    matrix<double>  indices;
    matrix<double>  indesi;
    matrix<double>  fluxes;
    try {
        indices::extract( fluxes, indices, indesi );
//        double h = test( indices );        mz = 1.0;
        std::cout << "M " << dc << ' ' << ut << " Normal " << std::fixed << std::setprecision(3) << mz << std::endl;
        std::ostringstream str;
        str << Date( floor(moment)) << ' ' << Time( 24*(moment-floor(moment)) );
        if( parameters::correction() ) indices::desaturation( indices );
        std::cout << indices::out_indices( str.str() ) << std::endl;
        std::cout << indices::out_fluxes( str.str() ) << std::endl;
        integrals::atmospheric( indices, indesi, mz );
        std::cout << integrals::out_atmosphere( str.str() ) << std::endl;
        restoration::calcn2( indices, mz, basis_functions );
//        std::cout << restoration::out_profile(  str.str() ) << "  " << h << std::endl;
        std::cout << restoration::out_profile(  str.str() ) << std::endl;
        std::cout << restoration::out_residual(  str.str() ) << std::endl;
        return 0;
    }
    catch( const std::exception& e  ) {
        std::cerr << "Processing fails: " << e.what() << ". The mode skipped " << std::endl;
        return 0;
    }
}

double  processing::object_Mz( Date& dc, Time& ut ) {
    return star.Mz( dc, ut );
}

void processing::object( std::vector<std::string>& items ) {
    static std::string current_type;
    static vector<double> innerDiam(nchan);
    static vector<double> outerDiam(nchan);
    static bool apertures_are_set = parameters::apertures( innerDiam, outerDiam );
    Spectral_data spd = parameters::spectral_data_files();
    int cdi = ( items.at(1).size() < 4 )? 3 : 2;
    std::string  ra(items.at(cdi+0)+' '+items.at(cdi+1)+' '+items.at(cdi+2));
    std::string dec(items.at(cdi+3)+' '+items.at(cdi+4)+' '+items.at(cdi+5));
    star = Coord<equ>( Hour(ra), Degree(dec), parameters::coordinates() );
    spectral_type = items.at(cdi+6);
    std::transform( spectral_type.begin(), spectral_type.end(), spectral_type.begin(), std::ptr_fun(::tolower) );
    if( spectral_type == current_type ) return;
    spd.weights += '/' + spectral_type + ".wf";
    spd.spectra += '/' + spectral_type + ".sp";
    basis_functions.calculation( spd, innerDiam, outerDiam, 9, 19.6 );
    integrals::coefficients( spectral_type, basis_functions );
    current_type = spectral_type;
}

double processing::test( matrix<double>& indices ) { // generate needed turbulence and indices
    test_functions.interpolate( basis_functions ) ;
    static int lfix = 80;
    loop++;
    loop =  loop%hs.size();
//    std::cerr << loop << " " <<  hs(loop) << std::endl;
//    vector<double> b = column( test_functions.weights(), loop );
    vector<double> e = scalar_vector<double>( hs.size(), 0 );
//    for( int i = e.size()-1; i > e.size()-12; i-- ) e(i) = 0;
    for( int i = 10; i < e.size()-12; i++ ) e(i) = (hs(i+1)-hs(i-1))/hs(i);
  
//    vector<double> f = column( test_functions.weights(), loop );
    vector<double> f = prod( test_functions.weights(), e );
    double mean = 1.0;
    for( unsigned int i = 0; i < test_functions.weight_functions_number(); i++ ) {
        mean = sum(column( indices, i ))/(double)indices.size1();
        column( indices, i ) *= 0.95*power*f(i)/mean;
        column( indices, i ) += scalar_vector<double>( indices.size1(), 0.05*power*f(i));
    }
//    secz = 1.0;
    return hs(loop);
}
