#include <lsp/nnls.h>
#include <lsp/cholesky_decomposition.h>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/triangular.hpp>
#include <iostream>
#include <iomanip>
#include <algorithm>
#include <numeric>
#include "profile.h"
#include "typesdef.h"
                                //    header
const double MAXCHI2 = 999.0;
const double JPOW_SEE = 0.6 ;
const double LPOW_SEE = -0.2;
const double KS = 1.73588e+07 ;

inline double seeing(double cn2, double secz, double leff ) {
    return KS*pow( cn2/secz, JPOW_SEE )*pow( leff, LPOW_SEE );
}

using namespace boost::numeric::ublas;

template<class T> int zero_elements( const vector<T>&x ) {
    int n = 0;
    for( unsigned int j = 0; j < x.size(); j++ ) n += ( x(j) == 0 )? 1 : 0;
    return n;
}

template<class T> vector<T> element_invert( const vector<T>&x ) {
    vector<T> y = scalar_vector<T>(x.size(),1);
    for( unsigned int j = 0; j < x.size(); j++ ) y(j) /= x(j);
    return y;
}

template<class T> vector<T> diag_elements( const symmetric_matrix<T, upper >&x ) {
    typename matrix<T>::size_type r = x.size1();
    slice s(0, 1, r );
    return matrix_vector_slice< const symmetric_matrix<T, upper > > ( x, s, s );
}

namespace restoration {
    std::string cn2method = "  X";
    vector<double> cn2;
    vector<double> ecn2;
    double chi2 = 0;
    double cn2seeing = 0;
    vector<double> userzcn2;
    vector<double> o_csi;
}

void restoration::setzcn2( const vector<double>& z ) {
    userzcn2 = z;
}

void restoration::covariance( matrix<double>& in,  vector<double>& mean, symmetric_matrix< double, upper >& cov ) {
    const double damper = 0.8;
    unsigned int n = in.size1();
    unsigned int k = in.size2();
    for( unsigned int i = 0; i < k; i++ ) mean(i) = sum(column( in, i ))/(double)n;
    for( unsigned int i = 0; i < k; i++) for( unsigned int j = i; j < k; j++) 
        cov(i,j) = inner_prod( column( in, i ), column( in, j ))/n - mean(i)*mean(j);
    for( unsigned int i = 0; i < k; i++) for( unsigned int j = i+1; j < k; j++)
        cov(i,j) = damper*cov(i,j);
    if( zero_elements( diag_elements( cov ) ) > 0 ) {
        std::ostringstream str;
        str << "Zero variances: " << n << "x" << k << " "  << mean << diag_elements( cov );
        throw std::runtime_error( str.str() );
    }
}

double restoration::nnls_solve( const matrix< double >& Wf, matrix<double>& inds,
       vector<double>& cn2, vector<double>& err, vector<double>& rsd ) {
    double ret = 0;
    unsigned int nw  =  Wf.size1();
    unsigned int ncn2 = Wf.size2();
    unsigned int nidx = inds.size1();
    assert( nw > 1 );
    assert( ncn2 > 1 );
    matrix< double > weights = identity_matrix< double >( nw );
    try{
        if( nidx < 2 ) {
            std::ostringstream str;
            str << "Too few measurements: " << nidx << "x" << inds.size2();
            throw std::runtime_error( str.str() );
        }
        vector<double> mean = zero_vector< double >( nw );
        symmetric_matrix< double, upper > covars( nw, nw );
        covariance( inds, mean, covars );
        lsp::cholesky_decomposition( covars, upper_tag() );             // covars  is broken now
        triangular_adaptor< symmetric_matrix< double, upper >, lower > U(covars);
        inplace_solve( U, weights, lower_tag());
        for( unsigned int i = 0; i < nw; i++ )
            if( weights(i,i) == 0 || isinf(weights(i,i)) || isnan(weights(i,i)) ) {
                std::ostringstream str;
                str << "Zero or infinity weights: " << diag_elements(weights);
                throw std::runtime_error( str.str() );
            }
        matrix< double > A = prod( weights, Wf );
        matrix< double > profile( nidx, ncn2 );
        vector< double > sc2( nidx );
        vector< double > e2 = zero_vector< double >( ncn2 );
        vector< double > r2 = zero_vector< double >( nw );
        cn2 = zero_vector< double >( ncn2 );
        for( unsigned int k = 0; k < nidx; k++ ) {
            vector< double > b = prod( weights, row( inds, k) );
            matrix< double > cov = zero_matrix<double>(ncn2, ncn2);
            lsp::nnls< matrix< double >, vector< double >  > nnls( A, b );
            vector< double > y;
            nnls.solve( y , cov );
            vector< double > r = prod( A, y ) - b;
            sc2( k ) = inner_prod( r, r );
            r2 += r;
            cn2 += y/(double)nidx;
            e2 += element_prod( y, y );
            row( profile, k ) = y;
        }
        err = ( e2/(double)nidx - element_prod( cn2, cn2 ))/(double)(nidx-1);
        rsd = element_prod( prod( U, r2/(double)nidx), apply_to_all( mean, functor::inv<double>() ));
        ret = sum( sc2 )/(double)nidx;
    }
    catch( const std::exception& err ){
        ret = -1.0;
        std::cerr << err.what() << std::endl;
	}
    return ret;
}

double restoration::calcn2( matrix<double>& ind, double secz, const Weights& basewf  ) {
    Weights wfs( basewf.apertures_number(), userzcn2*secz );
    wfs.interpolate( basewf ) ;
                                    // Solve the task by nnls method
    if( ( chi2 = restoration::nnls_solve( wfs.weights(), ind, cn2, ecn2, o_csi ) ) < 0 ) {
        cn2.clear();
        ecn2.clear();
    } else {
        cn2 /= secz;
        ecn2 /= secz;
    }
    chi2 = std::min( chi2, MAXCHI2 );
                                    // compute and return the seeing which corresponds to restored profile.
                                    // Zenith distance is set to 1 since Cn2[] is already corrected for it
    return cn2seeing = seeing( sum( cn2 ), 1, basewf.lambda_eff() );
}

std::string restoration::out_profile( const std::string& stamp ) {
    std::ostringstream str;
    str << "T " << stamp << cn2method << "  " << std::setw(2) << cn2.size() << "  " << std::setprecision(2)
        << std::fixed << chi2 << "  " << cn2seeing;
    for( unsigned int i = 0; i < cn2.size(); i++) 
        str << "  " << std::fixed << userzcn2(i) << std::scientific  << " " << cn2(i) << " " << sqrt(ecn2(i));
    return str.str();
}

std::string restoration::out_residual( const std::string& stamp ) {
    std::ostringstream str;
    str << "R " << stamp << cn2method << std::setprecision(3) << std::fixed;
    for( unsigned int i = 0; i < o_csi.size(); i++) str << "  " << o_csi(i);
    return str.str();
}

