#ifndef _WEIF_H
#define _WEIF_H

#include <string>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>

using namespace boost::numeric::ublas;

struct Spectral_data {
    std::string response;
    std::string weights;
    std::string spectra;
    std::string ccdresponse;
    Spectral_data( const std::string& r, const std::string& w, const std::string& s, const std::string& c)
        : response( r ), weights( w ), spectra( s ), ccdresponse(c) {}
};

class Spline;

typedef struct {
    double diameter1;
    double diameter2;
    double diameter3;
    double diameter4;
    double PIx2lambda0;
    Spline *re_sfun;
    Spline *im_sfun;
    double altitude;
} mass_params;

typedef struct {
    double aperture;
    double dimmbase;
    double PIx2lambda0;
    double phi;
    Spline *re_sfun;
    Spline *im_sfun;
    double altitude;
} dimm_params;

class Weights {
    unsigned int  mass_nchan;
    unsigned int  weights_number;
    vector<double> zgrid;           /**< vector of altitudes: z-grid [km] */
    matrix<double> ws;              /**< matrix of weights ws[aper][z] */
    vector<double> eps_inn;         /**< real inner diameters of apertures in [cm] */
    vector<double> eps_out;         /**< real outer diameters of apertures in [cm] */
    double leff;                    /**< effective wavelength of weight func. [mkm] */
    double lblue;                   /**< 50\% level wavelength at the blue side of SED [mkm] */
    double lred;                    /**< 50\% level wavelength at the red side of SED [mkm] */
    double dimm_aperture;
    double dimm_base;
public:
    Weights( int, const vector<double>& );
    ~Weights() {}
    const matrix<double>& weights() const { return ws; }
    const vector<double>& grid() const { return zgrid; }
    unsigned int apertures_number() const { return mass_nchan; }
    unsigned int weight_functions_number() const { return weights_number; }
    double lambda_eff() const { return leff ;}
    void interpolate( const Weights& refwf );
    void calculation( const Spectral_data&, const vector<double>&, const vector<double>&, double, double );
private:
    void write( const std::string& ) const ;
    void calc( vector<double>&, vector<double>&, vector<double>&, vector<double>& );
    int calcleff( vector<double>& lambda, vector<double>& edist );
    void photon_distr( vector<double>&, vector<double>&, vector<double>&, vector<double>& );
    void read_crv( const std::string&, vector<double>&, vector<double>& ) ;
    void spectral_band_fft( vector<double>&, vector<double>&, int, Spline*&, Spline*& ) ;
    vector<double> mass_weight( const vector<double>&, double, double, double, double, double, Spline*, Spline* );
    vector<double> dimm_weight( const vector<double>&, double, double, double, double, Spline*, Spline* );
};

namespace wf {
    template<class T> vector<T> grid_generate( unsigned int n, const T* z ) {
        const int z0 = 0; 
        const int zmax = 1;
        const int dz = 2;
        const int dzmin = 3;
        vector<T> h = scalar_vector<double>(1,z[z0]);
        if( n == 4 && z[zmax]> z[dz] && z[dzmin] != 0 ) {
            for( unsigned int i = 0; h(i) < z[zmax] ; i++ ) {
                h.resize(h.size()+1);
                h(i+1) = h(i) + std::max(0.5*h(i)*z[dz],z[dzmin]);
            }
        }
        if( n == 4 && z[zmax] > z[dz] && z[dzmin] == 0 ) {
            for( unsigned int i = 0; h(i) < z[zmax] ; i++ ) {
                h.resize(h.size()+1);
                h(i+1) = h(i) + z[dz];
            }
        }
        if( n > 1 && z[zmax]< z[dz] && z[dz] < z[dzmin] ) {
            h.resize(n);
            for( unsigned int i = 0; i < n ; i++ ) h(i) = z[i];
        }
        return h;
    }
}

#endif
